###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2014'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

import os
import errno
import sys
import logging
import ntpath
import re
import gzip

from .exceptions import BioLibFileNotFound, BioLibDirNotFound, BioLibIOException


def is_float(s):
    # type: (str) -> bool
    """Check if a string can be converted to a float.
    Parameters
    ----------
    s : str
        String to evaluate.
    Returns
    -------
    boolean
        True if string can be converted, else False.
    """

    try:
        float(s)
    except ValueError:
        return False

    return True


def check_file_exists(input_file):
    """Check if file exists."""
    if not os.path.exists(input_file) or not os.path.isfile(input_file):
        logger = logging.getLogger('timestamp')
        logger.error('Input file does not exist: ' + input_file + '\n')
        raise BioLibFileNotFound('Input file does not exist: ' + input_file + '\n')
    return True


def check_dir_exists(input_dir):
    """Check if directory exists."""
    if not os.path.exists(input_dir) or not os.path.isdir(input_dir):
        logger = logging.getLogger('timestamp')
        logger.error('Input directory does not exist: ' + input_dir + '\n')
        raise BioLibDirNotFound('Input directory does not exist: ' + input_dir + '\n')
    return True


def make_sure_path_exists(path):
    """Create directory if it does not exist."""

    if not path:
        # lack of a path qualifier is acceptable as this
        # simply specifies the current directory
        return True

    elif os.path.isdir(path):
        return True

    try:
        os.makedirs(path)
        return True
    except OSError:
            logger = logging.getLogger('timestamp')
            logger.error('Specified path could not be created: ' + path + '\n')
            raise BioLibIOException('Specified path could not be created: ' + path + '\n')


def remove_extension(filename, extension=None):
    # type: (str, str) -> str
    """Remove extension from filename.
    A specific extension can be specified, otherwise
    the extension is taken as all characters after the
    last period.
    """
    f = ntpath.basename(filename)

    if extension and f.endswith(extension):
        f = f[0:f.rfind(extension)]
    else:
        f = os.path.splitext(f)[0]

    if f[-1] == '.':
        f = f[0:-1]

    return f



# import prodigal
import math
import time
import random
import os
import hashlib

from itertools import islice
from biolib_lite.seq_io import read_fasta
from biolib_lite.common import remove_extension


##################################################
############MISC UTILITIES########################
##################################################

def add_ncbi_prefix(refname):
    if refname.startswith("GCF_"):
        return "RS_" + refname
    elif refname.startswith("GCA_"):
        return "GB_" + refname
    else:
        return refname


def splitchunks(d, n):
    chunksize = int(math.ceil(len(d) / float(n)))
    it = iter(d)
    for _ in range(0, len(d), chunksize):
        yield {k: d[k] for k in islice(it, chunksize)}


def splitchunks_list(l, n):
    """Yield successive n-sized chunks from l."""
    chunksize = int(math.ceil(len(l) / float(n)))
    for i in range(0, len(l), chunksize):
        yield l[i:i + chunksize]


def generateTempTableName():
    rng = random.SystemRandom()
    suffix = ''
    for _ in range(0, 10):
        suffix += rng.choice(
            'abcefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')
    return "TEMP" + suffix + str(int(time.time()))


def merge_two_dicts(x, y):
    '''Given two dicts, merge them into a new dict as a shallow copy.'''
    z = x.copy()
    z.update(y)
    return z


def sha256(input_file):
    """Determine SHA256 hash for file.
    Parameters
    ----------
    input_file : str
        Name of file.
    Returns
    -------
    str
        SHA256 hash.
    """

    BLOCKSIZE = 65536
    hasher = hashlib.sha1()
    with open(input_file, 'rb') as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)

    return hasher.hexdigest()


def symlink_f(src, dst, force=True):
    """Create a symbolic link pointing to src named dst.

    Parameters
    ----------
    src : str
        The source file.
    dst : str
        The destination file.
    force : bool
        Overwrite any file found with the same name as dst.

    """
    if force and os.path.isfile(dst):
        os.remove(dst)
    os.symlink(src, dst)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import re
import sys
from collections import deque

from .HMMResults import HMMResults
from .HMMSequence import HMMSequence
from .HMMUnit import HMMUnit


class HMMResultsIO(object):
    """
    This class has been adapted from the Perl module written by Genome Research Ltd.
    Perl authors: Rob Finn (rdf@sanger.ac.uk), John Tate (jt6@sanger.ac.uk)
    Perl version: $Id: HMMResultsIO.pm,v 1.2 2009-12-01 15:42:20 jt6 Exp $
    Python authors: Aaron Mussig (a.mussig@uq.edu.au)
    """

    def __init__(self):
        
        self.align = 0
        self.outfile = 'OUTPUT'
        self.pfamout = 'PFAMOUT'
        self.scores = 'scores'

        # Python adaptation: Pre-compile regex patterns used in _readHeader
        self.re_header_end = re.compile(r'^Scores for complete')
        self.re_hmm_name = re.compile(r'^# query HMM file:\s+(\S+)')
        self.re_seq_db = re.compile(r'^# target sequence database:\s+(\S+)')
        self.re_this_file = re.compile(r'^output directed to file:\s+(\S+)')
        self.re_seed_name__hmm_length = re.compile(r'^Query:\s+(\S+)\s+\[M\=(\d+)\]')
        self.re_seq_name__seq_length = re.compile(r'^Query:\s+(\S+)\s+\[L\=(\d+)\]')
        self.re_e_value_Thr = re.compile(r'^sequence E-value threshold: <= (\d+)')
        self.re_rand_seed_num = re.compile(r'^# Random generator seed:      (\d+)')
        self.re_descr = re.compile(r'^Description:\s+(.*)')
        self.re_program = re.compile(r'^# (phmmer|hmmsearch|hmmscan|jackhmmer)')
        self.re_header_skip = re.compile(r'(^#)|(^$)')
        self.re_header_skip_2 = re.compile(r'^Accession')
        self.re_eof = re.compile(r'^\[ok\]')

        # Python adaptation: Pre-compile the regex used in parsing the sequence hits.
        self.re_seq_hit_1 = re.compile(r'^Domain annotation for each [sequence|model]')
        self.re_seq_hit_2 = re.compile(r'^Domain and alignment annotation for each [sequence|model]')
        self.re_seq_hit_3 = re.compile(r'^\s+(E-value|---)')
        self.re_seq_hit_4 = re.compile(r'^$')
        self.re_seq_hit_5 = re.compile(r'No hits detected that satisfy reporting thresholds')
        self.re_seq_hit_6 = re.compile(r'\s+')

        # Python adaptation: Pre-compile the regex used in parsing the domain hits.
        self.re_dom_1 = re.compile(r'^Internal')
        self.re_dom_2 = re.compile(r'\>\>\s+(\S+)')

        # Python adaptation: Pre-compile the regex used in parsing the unit data.
        self.re_unitd_1 = re.compile(r'^[(\/\/|Internal)]')
        self.re_unitd_2 = re.compile(r'^\>\>\s+(\S+)')
        self.re_unitd_3 = re.compile(r'^\s+Alignments for each domain:')
        self.re_unitd_4 = re.compile(r'^\s+(#\s+score|---)')
        self.re_unitd_5 = re.compile(r'^$')
        self.re_unitd_6 = re.compile(r'^\s+\d+\s+')
        self.re_unitd_7 = re.compile(r'\s+')
        self.re_unitd_8 = re.compile(r'^\s+\[No individual domains')
        self.re_unitd_9 = re.compile(r'^\s+([x\.]+)\s+RF$')
        self.re_unitd_10 = re.compile(r'^\s+([0-9\*\.]+)\s+PP$')
        self.re_unitd_11 = re.compile(r'^\s+(\S+)\s+CS$')
        self.re_unitd_12 = re.compile(r'^\s+==\s+domain\s+(\d+)')
        self.re_unitd_13 = re.compile(r'^\s+(.*)\s*$')  # Modified from Perl version as no newline break in iterator.
        self.re_unitd_14 = re.compile(r'^$')
        self.re_unitd_15 = re.compile(r'^[(\/\/|Internal)]')
        self.re_unitd_16 = re.compile(r'^\>\>\s+(\S+)')
        
        # Python adaptation: Pre-compile the regex used in parsing the footer.
        self.re_footer_1 = re.compile(r'\/\/')

        # Python adaptation: Pre-compile the regex used in the ascii_output function
        self._re_ascii_1 = re.compile(r'Pfam\-B')

    def parseMultiHMMER3(self, filename):

        fh = deque(filename.splitlines())

        hmmResAll = list()
        program = None

        while len(fh) > 0:
            hmmRes = HMMResults()

            eof = self._readHeader(fh, hmmRes)
            if eof:
                break

            hmmResAll.append(hmmRes)

            if hmmRes.program:
                program = hmmRes.program
            else:
                hmmRes.program = program

            self._readSeqHits(fh, hmmRes)
            self._readUnitHits(fh, hmmRes)
            self._readFooter(fh, hmmRes)

        return hmmResAll

    def _readHeader(self, fh, hmmRes):
        """
        Reads the header section from a HMMER3 hmmsearch
        """

        while len(fh) > 0:
            hs = fh.popleft()

            # Python adaptation: Execute all regex searches which extract information.
            hmm_name = self.re_hmm_name.search(hs)
            seq_db = self.re_seq_db.search(hs)
            this_file = self.re_this_file.search(hs)
            seed_name__hmm_length = self.re_seed_name__hmm_length.search(hs)
            seq_name__seq_length = self.re_seq_name__seq_length.search(hs)
            e_value_Thr = self.re_e_value_Thr.search(hs)
            rand_seed_num = self.re_rand_seed_num.search(hs)
            descr = self.re_descr.search(hs)
            program = self.re_program.search(hs)

            if self.re_header_end.search(hs):
                break
            elif hmm_name:
                hmmRes.hmmName = hmm_name.group(1)
            elif seq_db:
                hmmRes.seqDB = seq_db.group(1)
            elif this_file:
                hmmRes.thisFile = this_file.group(1)
            elif seed_name__hmm_length:
                hmmRes.seedName = seed_name__hmm_length.group(1)
                hmmRes.hmmLength = seed_name__hmm_length.group(2)
            elif seq_name__seq_length:
                hmmRes.seqName = seq_name__seq_length.group(1)
                hmmRes.seqLength = seq_name__seq_length.group(2)
            elif e_value_Thr:
                hmmRes.evalueThr = e_value_Thr.group(1)
            elif rand_seed_num:
                hmmRes.randSeedNum = rand_seed_num.group(1)
            elif descr:
                hmmRes.description = descr.group(1)
            elif program:
                hmmRes.program = program.group(1)
            elif self.re_header_skip.search(hs):
                continue
            elif self.re_header_skip_2.search(hs):
                continue
            elif self.re_eof.search(hs):
                return True
            else:
                sys.exit("Failed to parse hmmsearch results in header section\n")

    def _readSeqHits(self, fh, hmmRes):
        """
        Reads the sequence hits from a HMMER3 hmmsearch
        """

        while len(fh) > 0:
            hs = fh.popleft()

            if self.re_seq_hit_1.search(hs):
                break
            elif self.re_seq_hit_2.search(hs):
                break
            elif self.re_seq_hit_3.search(hs):
                continue
            elif self.re_seq_hit_4.search(hs):
                continue
            elif self.re_seq_hit_5.search(hs):
                continue
            else:

                # Assume that we have a sequence match
                re_6 = self.re_seq_hit_6.split(hs)

                # CHeck if it is what is expected
                if len(re_6) < 10:
                    sys.exit('Expected at least 10 pieces of data.\n')

                descr = ' '.join(re_6[10:])

                hmmSeq = HMMSequence()
                hmmSeq.evalue = float(re_6[1])
                hmmSeq.bits = float(re_6[2])
                hmmSeq.bias = float(re_6[3])
                hmmSeq.exp = float(re_6[7])
                hmmSeq.numberHits = int(re_6[8])
                hmmSeq.name = re_6[9]
                hmmSeq.desc = descr if descr else '-'
                hmmRes.addHMMSeq(hmmSeq)

    def _readUnitHits(self, fh, hmmRes):
        """
        Reads the unit (domain) hits from a HMMER3 hmmsearch
        """

        if hmmRes.eof:
            return

        # Parse the domain hits section
        while len(fh) > 0:
            hs = fh.popleft()

            # Perform the regex searches
            re_dom_2 = self.re_dom_2.search(hs)

            if self.re_dom_1.search(hs):
                break

            elif re_dom_2:
                seqId = re_dom_2.group(1)
                self._readUnitData(seqId, fh, hmmRes)

                if hmmRes.eof:
                    return

    def _readUnitData(self, seqId, fh, hmmRes):

        if hmmRes.eof:
            return

        hmmName = hmmRes.seedName
        seqName = hmmRes.seqName

        units = list()
        align = True
        recurse = False
        eof = False
        nextSeqId = None

        # Parse the domain hits section
        while len(fh) > 0:
            hs = fh.popleft()

            # Run the regex searches which generate output
            unitd_2 = self.re_unitd_2.search(hs)

            if self.re_unitd_1.search(hs):
                align = False
                recurse = False
                eof = True
                break

            elif unitd_2:
                nextSeqId = unitd_2.group(1)
                align = False
                recurse = True
                break

            elif self.re_unitd_3.search(hs):
                align = True
                recurse = False
                break

            elif self.re_unitd_4.search(hs):
                # Two human readable lines
                continue

            elif self.re_unitd_5.search(hs):
                # blank line
                continue

            elif self.re_unitd_6.search(hs):
                dMatch = self.re_unitd_7.split(hs)
                if len(dMatch) != 17:
                    sys.exit('Expected 16 elements of data.')

                hmmUnit = HMMUnit()
                hmmUnit.name = seqId
                hmmUnit.domain = dMatch[1]
                hmmUnit.bits = float(dMatch[3])
                hmmUnit.bias = float(dMatch[4])
                hmmUnit.domEvalue = float(dMatch[5])
                hmmUnit.evalue = float(dMatch[6])
                hmmUnit.hmmFrom = int(dMatch[7])
                hmmUnit.hmmTo = int(dMatch[8])
                hmmUnit.seqFrom = int(dMatch[10])
                hmmUnit.seqTo = int(dMatch[11])
                hmmUnit.envFrom = int(dMatch[13])
                hmmUnit.envTo = int(dMatch[14])
                hmmUnit.aliAcc = float(dMatch[16])

                units.append(hmmUnit)
                continue

            elif self.re_unitd_8.search(hs):
                align = False
                continue

            else:
                sys.exit('Did not parse line %s' % hs)

        '''
        #  == domain 1    score: 244.0 bits;  conditional E-value: 9.5e-76
        #      SEED   1 medrlallkaisasakdlvalaasrGaksipspvkttavkfdplptPdldalrtrlkeaklPakaiksalsayekaCarWrsdleeafdktaksvsPanlhllealrirlyteqvekWlvqvlevaerWkaemekqrahiaatmgp 146
        #               m+++la+l++isa+akd++ala+srGa+++ +p++tt+++fd+l++P+ld++rtrl+ea+lP+kaik++lsaye+aCarW++dleeafd+ta+s+sP+n+++l++lr+rly+eqv+kWl++vl+v+erWkaemekqrahi+atmgp
        #  P37935.1   1 MAELLACLQSISAHAKDMMALARSRGATGS-RPTPTTLPHFDELLPPNLDFVRTRLQEARLPPKAIKGTLSAYESACARWKHDLEEAFDRTAHSISPHNFQRLAQLRTRLYVEQVQKWLYEVLQVPERWKAEMEKQRAHINATMGP 145
        #               899***************************.******************************************************************************************************************8 PP
        #
        # OR....
        #
        #  == domain 1    score: 27.6 bits;  conditional E-value: 7.4e-10
        #   PF00018  17 LsfkkGdvitvleksee.eWwkaelkdg.keGlvPsnYvep 55 
        #               L++++Gd+++++++++e++Ww++++++++++G++P+n+v+p
        #  P15498.4 617 LRLNPGDIVELTKAEAEqNWWEGRNTSTnEIGWFPCNRVKP 657
        #               7899**********9999*******************9987 PP
        '''

        if align:

            # Specifically for python
            pattern1 = None
            pattern2 = None

            if hmmName and hmmRes.program == 'hmmsearch':

                pattern1 = re.compile(r'^\s+%s\s+\d+\s+(\S+)\s+\d+' % hmmName)
                seqId = re.sub('(\W)', r'\\\1', seqId)
                # $id =~ s/\|/\\|/g;  #Escape '|', '[' and ']' characters
                # $id =~ s/\[/\\[/g;
                # $id =~ s/\]/\\]/g;
                pattern2 = re.compile(r'^\s+%s\s+\d+\s+(\S+)\s+\d+' % seqId)

            elif seqName and hmmRes.program == 'hmmscan':
                tmpSeqName = seqName
                tmpSeqName = re.sub('(\W)', r'\\\1', tmpSeqName)

                pattern1 = re.compile(r'^\s+%s\s+\d+\s+(\S+)\s+\d+' % seqId)
                pattern2 = re.compile(r'^\s+%s\s+\d+\s+(\S+)\s+\d+' % tmpSeqName)

            elif seqName and (hmmRes.program == 'phmmer' or hmmRes.program == 'jackhmmer'):
                sys.exit("seqName and (hmmRes.program == 'phmmer' or hmmRes.program == 'jackhmmer' is not implemented.")

            recurse = False
            matchNo = None
            hmmlen = 0

            while len(fh) > 0:
                hs = fh.popleft()

                # Run a search for each of the patterns.
                pattern1_res = pattern1.search(hs)
                pattern2_res = pattern2.search(hs)
                re_unitd_9_res = self.re_unitd_9.search(hs)
                re_unitd_10_res = self.re_unitd_10.search(hs)
                re_unitd_11_res = self.re_unitd_11.search(hs)
                re_unitd_12_res = self.re_unitd_12.search(hs)
                re_unitd_13_res = self.re_unitd_13.search(hs)
                re_unitd_14_res = self.re_unitd_14.search(hs)
                re_unitd_15_res = self.re_unitd_15.search(hs)
                re_unitd_16_res = self.re_unitd_16.search(hs)

                if pattern1_res:
                    dict_hmmalign = units[matchNo - 1].hmmalign
                    if 'hmm' in dict_hmmalign:
                        dict_hmmalign['hmm'] += pattern1_res.group(1)
                    else:
                        dict_hmmalign['hmm'] = pattern1_res.group(1)

                    hmmlen = len(pattern1_res.group(1))

                elif pattern2_res:
                    dict_hmmalign = units[matchNo - 1].hmmalign
                    if 'seq' in dict_hmmalign:
                        dict_hmmalign['seq'] += pattern2_res.group(1)
                    else:
                        dict_hmmalign['seq'] = pattern2_res.group(1)

                # ^\s+([x\.]+)\s+RF$
                elif re_unitd_9_res:
                    rf = re_unitd_9_res.group(1)
                    dict_hmmalign = units[matchNo - 1].hmmalign
                    if 'rf' in dict_hmmalign:
                        dict_hmmalign['rf'] += rf
                    else:
                        dict_hmmalign['rf'] = rf

                # ^\s+([0-9\*\.]+)\s+PP$
                elif re_unitd_10_res:
                    pp = re_unitd_10_res.group(1)

                    dict_hmmalign = units[matchNo - 1].hmmalign
                    if 'pp' in dict_hmmalign:
                        dict_hmmalign['pp'] += pp
                    else:
                        dict_hmmalign['pp'] = pp

                # ^\s+(\S+)\s+CS$
                elif re_unitd_11_res:
                    cs = re_unitd_11_res.group(1)
                    dict_hmmalign = units[matchNo - 1].hmmalign
                    if 'cs' in dict_hmmalign:
                        dict_hmmalign['cs'] += cs
                    else:
                        dict_hmmalign['cs'] = cs

                # ^\s+==\s+domain\s+(\d+)
                elif re_unitd_12_res:
                    matchNo = int(re_unitd_12_res.group(1))

                # ^\s+(.*)\s+$
                elif re_unitd_13_res:
                    hs = hs.rstrip()
                    m1 = hs[-hmmlen:]

                # ^$
                elif re_unitd_14_res:
                    continue

                # ^[(\/\/|Internal)]
                elif re_unitd_15_res:
                    align = False
                    recurse = False
                    eof = True
                    break

                # ^\>\>\s+(\S+)
                elif re_unitd_16_res:
                    nextSeqId = re_unitd_16_res.group(1)
                    recurse = True
                    break

                else:
                    sys.exit('Did not parse %s in units' % hs)

        # foreach my u (@units)
        for u in units:
            hmmRes.addHMMUnit(u)

        hmmRes.eof = eof

        if (recurse and nextSeqId):
            self._readUnitData(nextSeqId, fh, hmmRes)

        return

    def _readFooter(self, fh, hmmRes):

        # We are going to parse something like this!

        #  Internal pipeline statistics summary:
        # -------------------------------------
        # Query sequence(s):                         1  (360 residues)
        # Target model(s):                           7  (836 nodes)
        # Passed MSV filter:                         2  (0.285714); expected 0.1 (0.02)
        # Passed Vit filter:                         1  (0.142857); expected 0.0 (0.001)
        # Passed Fwd filter:                         1  (0.142857); expected 0.0 (1e-05)
        # Initial search space (Z):                  7  [actual number of targets]
        # Domain search space  (domZ):               1  [number of targets reported over threshold]
        ## CPU time: 0.00u 0.00s 00:00:00.00 Elapsed: 00:00:00
        ## Mc/sec: inf
        # //

        while len(fh) > 0:
            hs = fh.popleft()

            if self.re_footer_1.search(hs):
                break

    def write_ascii_out(self, HMMResults, fh, scanData, e_seq, e_dom, b_seq, b_dom):

        if not scanData._max_seqname or scanData._max_seqname < 1:
            scanData._max_seqname = 20

        ga = None
        if e_seq or e_dom:
            sys.exit('FATAL: e_seq/e_dom functionality has not been implemented in Python')
        elif b_seq or b_dom:
            sys.exit('FATAL: b_seq/b_dom functionality has not been implemented in Python')
        else:
            ga = True

        for unit in sorted(HMMResults.units, key=lambda x: x.seqFrom, reverse=False):

            # Pfam\-B
            if self.re_footer_1.search(unit.name):
                sys.exit('FATAL: Pfam-B functionality has not been implemented in Python')

            else:

                # Filter results based on thresholds
                if ga:
                    if not unit.sig:
                        continue
                if e_seq:
                    sys.exit('FATAL: e_seq functionality has not been implemented in Python')
                if b_seq:
                    sys.exit('FATAL: b_seq functionality has not been implemented in Python')

                clan = "No_clan" if unit.name not in scanData._clanmap else scanData._clanmap[unit.name]

                seq_name_padded = HMMResults.seqName + ' ' * (scanData._max_seqname - len(HMMResults.seqName))
                fh.write('%s %6d %6d %6d %6d %-11s %-16s %7s %5d %5d %5d %8s %9s %3d %-8s ' %
                      (seq_name_padded,
                       unit.seqFrom,
                       unit.seqTo,
                       unit.envFrom,
                       unit.envTo,
                       scanData._accmap[unit.name],
                       unit.name,
                       scanData._type[unit.name],
                       unit.hmmFrom,
                       unit.hmmTo,
                       scanData._model_len[unit.name],
                       unit.bits,
                       unit.evalue,
                       unit.sig,
                       clan))

                fh.write('\n')

                if scanData._translate:
                    sys.exit('FATAL: translate functionality has not been implemented in Python')

            if scanData._align:
                sys.exit('FATAL: Align functionality has not been implemented in Python')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from __future__ import print_function

import os
import logging
import sys
import shutil

from gtdbtk.tools import symlink_f
from markers import Markers
from classify import Classify
from misc import Misc
from reroot_tree import RerootTree
import config.config as Config
from gtdbtk.config.output import *

from biolib_lite.common import (check_dir_exists,
                                check_file_exists,
                                make_sure_path_exists,
                                remove_extension)
from biolib_lite.taxonomy import Taxonomy
from biolib_lite.execute import check_dependencies
from gtdbtk.exceptions import *


class OptionsParser(object):

    def __init__(self, version):
        """Initialization."""
        self.logger = logging.getLogger('timestamp')
        self.version = version

        self.logger.warning(
            "Results are still being validated and taxonomic assignments may be incorrect! Use at your own risk!")

        self._check_package_compatibility()

    def _check_package_compatibility(self):
        """Check that GTDB-Tk is using the most up-to-date reference package."""
        pkg_ver = float(Config.VERSION_DATA.replace('r', ''))
        min_ver = float(Config.MIN_REF_DATA_VERSION.replace('r', ''))
        self.logger.info('Using GTDB-Tk reference data version {}: {}'
                         .format(Config.VERSION_DATA, Config.GENERIC_PATH))
        if pkg_ver < min_ver:
            self.logger.warning('You are not using the reference data intended '
                                'for this release: {}'.format(Config.MIN_REF_DATA_VERSION))

    def _verify_genome_id(self, genome_id):
        """Ensure genome ID will be valid in Newick tree."""

        invalid_chars = set('()[],;=')
        if any((c in invalid_chars) for c in genome_id):
            self.logger.error('Invalid genome ID: %s' % genome_id)
            self.logger.error('The following characters are invalid: %s' % ' '.join(invalid_chars))
            raise GenomeNameInvalid
        return True

    def _genomes_to_process(self, genome_dir, batchfile, extension):
        """Get genomes to process.

        Parameters
        ----------
        genome_dir : str
          Directory containing genomes.
        batchfile : str
          File describing genomes.
        extension : str
          Extension of files to process.

        Returns
        -------
        genomic_files : d[genome_id] -> FASTA file
            Map of genomes to their genomic FASTA files.
        """

        genomic_files = {}
        if genome_dir:
            for f in os.listdir(genome_dir):
                if f.endswith(extension):
                    genome_id = remove_extension(f)
                    genomic_files[genome_id] = os.path.join(genome_dir, f)

        elif batchfile:
            for line_no, line in enumerate(open(batchfile, "rb")):
                line_split = line.strip().split("\t")
                if line_split[0] == '':
                    continue  # blank line

                if len(line_split) != 2:
                    self.logger.error('Batch file must contain exactly 2 columns.')
                    raise GenomeBatchfileMalformed

                genome_file, genome_id = line_split
                self._verify_genome_id(genome_id)

                if genome_file is None or genome_file == '':
                    self.logger.error('Missing genome file on line %d.' % (line_no + 1))
                    raise GenomeBatchfileMalformed
                elif genome_id is None or genome_id == '':
                    self.logger.error('Missing genome ID on line %d.' % (line_no + 1))
                    raise GenomeBatchfileMalformed
                elif genome_id in genomic_files:
                    self.logger.error('Genome ID %s appear multiple times.' % genome_id)
                    raise GenomeBatchfileMalformed
                if genome_file in genomic_files.values():
                    self.logger.warning('Genome file appears multiple times: %s' % genome_file)

                genomic_files[genome_id] = genome_file

        for genome_key in genomic_files.iterkeys():
            if genome_key.startswith("RS_") or genome_key.startswith("GB_") or genome_key.startswith("UBA"):
                self.logger.error(
                    "Submitted genomes start with the same prefix (RS_,GB_,UBA) as reference genomes in GTDB-Tk. This will cause issues for downstream analysis.")
                raise GenomeNameInvalid

        if len(genomic_files) == 0:
            if genome_dir:
                self.logger.error(
                    'No genomes found in directory: %s. Check the --extension flag used to identify genomes.' % genome_dir)
            else:
                self.logger.error(
                    'No genomes found in batch file: %s. Please check the format of this file.' % batchfile)
            raise NoGenomesFound

        return genomic_files

    def _marker_set_id(self, bac120_ms, ar122_ms, rps23_ms):
        """Get unique identifier for marker set."""

        if bac120_ms:
            marker_set_id = "bac120"
        elif ar122_ms:
            marker_set_id = "ar122"
        elif rps23_ms:
            marker_set_id = "rps23"

        return marker_set_id

    def identify(self, options):
        """Identify marker genes in genomes."""

        if options.genome_dir:
            check_dir_exists(options.genome_dir)

        if options.batchfile:
            check_file_exists(options.batchfile)

        make_sure_path_exists(options.out_dir)

        genomes = self._genomes_to_process(
            options.genome_dir, options.batchfile, options.extension)

        markers = Markers(options.cpus)
        markers.identify(genomes,
                         options.out_dir,
                         options.prefix,
                         options.force)

        self.logger.info('Done.')

    def align(self, options):
        """Create MSA from marker genes."""

        check_dir_exists(options.identify_dir)
        make_sure_path_exists(options.out_dir)

        if not hasattr(options, 'outgroup_taxon'):
            options.outgroup_taxon = None

        markers = Markers(options.cpus)
        markers.align(options.identify_dir,
                      options.skip_gtdb_refs,
                      options.taxa_filter,
                      options.min_perc_aa,
                      options.custom_msa_filters,
                      options.skip_trimming,
                      options.rnd_seed,
                      options.cols_per_gene,
                      options.min_consensus,
                      options.max_consensus,
                      options.min_perc_taxa,
                      options.out_dir,
                      options.prefix,
                      options.outgroup_taxon)

        self.logger.info('Done.')

    def infer(self, options):
        """Infer tree from MSA."""

        check_file_exists(options.msa_file)
        make_sure_path_exists(options.out_dir)

        if options.cpus > 1:
            check_dependencies(['FastTreeMP'])
            os.environ['OMP_NUM_THREADS'] = '%d' % options.cpus
        else:
            check_dependencies(['FastTree'])

        if hasattr(options, 'suffix'):
            output_tree = os.path.join(options.out_dir,
                                       PATH_MARKER_UNROOTED_TREE.format(prefix=options.prefix, marker=options.suffix))
            tree_log = os.path.join(options.out_dir,
                                    PATH_MARKER_TREE_LOG.format(prefix=options.prefix, marker=options.suffix))
            fasttree_log = os.path.join(options.out_dir,
                                        PATH_MARKER_FASTTREE_LOG.format(prefix=options.prefix, marker=options.suffix))
        else:
            output_tree = os.path.join(options.out_dir, PATH_UNROOTED_TREE.format(prefix=options.prefix))
            tree_log = os.path.join(options.out_dir, PATH_TREE_LOG.format(prefix=options.prefix))
            fasttree_log = os.path.join(options.out_dir, PATH_FASTTREE_LOG.format(prefix=options.prefix))

        make_sure_path_exists(os.path.dirname(output_tree))
        make_sure_path_exists(os.path.dirname(tree_log))
        make_sure_path_exists(os.path.dirname(fasttree_log))

        if options.prot_model == 'JTT':
            model_str = ''
        elif options.prot_model == 'WAG':
            model_str = ' -wag'
        elif options.prot_model == 'LG':
            model_str = ' -lg'

        support_str = ''
        if options.no_support:
            support_str = ' -nosupport'

        gamma_str = ' -gamma'
        gamma_str_info = '+GAMMA'
        if options.no_gamma:
            gamma_str = ''
            gamma_str_info = ''

        self.logger.info(
            'Inferring tree with FastTree using {}.'.format(options.prot_model, gamma_str_info))

        cmd = '-quiet%s%s%s -log %s %s > %s 2> %s' % (support_str,
                                                      model_str,
                                                      gamma_str,
                                                      tree_log,
                                                      options.msa_file,
                                                      output_tree,
                                                      fasttree_log)
        if options.cpus > 1:
            cmd = 'FastTreeMP ' + cmd
        else:
            cmd = 'FastTree ' + cmd
        os.system(cmd)

        self.logger.info('Done.')

    def run_test(self, options):
        """Run test of classify workflow."""

        make_sure_path_exists(options.out_dir)

        output_dir = os.path.join(options.out_dir, 'output')
        genome_test_dir = os.path.join(options.out_dir, 'genomes')
        if os.path.exists(genome_test_dir):
            self.logger.error('Test directory {} already exists. Test must be run with a new directory.'.format(
                                genome_test_dir))
            sys.exit(-1)

        current_path = os.path.dirname(os.path.realpath(__file__))
        input_dir = os.path.join(current_path, 'tests', 'data', 'genomes')

        shutil.copytree(input_dir, genome_test_dir)

        cmd = 'gtdbtk classify_wf --genome_dir {} --out_dir {} --cpus {}'.format(
            genome_test_dir, output_dir, options.cpus)
        print("Command:")
        print(cmd)
        os.system(cmd)
        summary_file = os.path.join(output_dir, PATH_AR122_SUMMARY_OUT.format(prefix='gtdbtk'))

        if not os.path.exists(summary_file):
            print("{} is missing.\nTest has failed.".format(summary_file))
            sys.exit(-1)

        self.logger.info('Test has successfully finished.')

    def classify(self, options):
        """Determine taxonomic classification of genomes."""

        check_dir_exists(options.align_dir)
        make_sure_path_exists(options.out_dir)
        if options.scratch_dir:
            make_sure_path_exists(options.scratch_dir)

        genomes = self._genomes_to_process(
            options.genome_dir, options.batchfile, options.extension)

        classify = Classify(options.cpus)
        classify.run(genomes,
                     options.align_dir,
                     options.out_dir,
                     options.prefix,
                     options.scratch_dir,
                     options.recalculate_red,
                     options.debug)

        self.logger.info('Done.')

    def trim_msa(self, options):
        """ Trim an untrimmed archaea or bacterial MSA file."""
        if options.reference_mask in ['bac', 'arc']:
            mask_type = "reference"
            mask_id = options.reference_mask
        else:
            mask_type = "file"
            mask_id = options.mask_file
        misc = Misc()
        misc.trim_msa(options.untrimmed_msa, mask_type,
                      mask_id, options.output)
        self.logger.info('Done.')

    def export_msa(self, options):
        """Export the untrimmed archaeal or bacterial MSA file."""
        misc = Misc()
        misc.export_msa(options.domain, options.output)

        self.logger.info('Done.')

    def root(self, options):
        """Root tree using outgroup."""
        self.logger.warning("Tree rooting is still under development!")

        check_file_exists(options.input_tree)

        if options.custom_taxonomy_file:
            check_file_exists(options.custom_taxonomy_file)
            taxonomy = Taxonomy().read(options.custom_taxonomy_file)
        else:
            taxonomy = Taxonomy().read(Config.TAXONOMY_FILE)

        self.logger.info('Identifying genomes from the specified outgroup.')
        outgroup = set()
        for genome_id, taxa in taxonomy.iteritems():
            if options.outgroup_taxon in taxa:
                outgroup.add(genome_id)

        reroot = RerootTree()
        reroot.root_with_outgroup(options.input_tree,
                                  options.output_tree,
                                  outgroup)

        # Symlink to the tree summary file
        if options.suffix == 'bac120':
            symlink_f(PATH_BAC120_ROOTED_TREE.format(prefix=options.prefix),
                       os.path.join(options.out_dir, os.path.basename(PATH_AR122_ROOTED_TREE.format(prefix=options.prefix))))
        elif options.suffix == 'ar122':
            symlink_f(PATH_AR122_ROOTED_TREE.format(prefix=options.prefix),
                       os.path.join(options.out_dir, os.path.basename(PATH_AR122_ROOTED_TREE.format(prefix=options.prefix))))
        else:
            self.logger.error('There was an error determining the marker set.')
            raise GenomeMarkerSetUnknown

        self.logger.info('Done.')

    def check_install(self):
        """ Verify all GTDB-Tk data files are present."""
        self.logger.info("Running install verification")
        misc = Misc()
        misc.check_install()
        self.logger.info('Done.')

    def decorate(self, options):
        """Decorate tree with GTDB taxonomy."""

        check_file_exists(options.input_tree)

        # Config.TAXONOMY_FILE
        self.logger.warning('DECORATE NOT YET IMPLEMENTED!')
        self.logger.info('Done.')

    def parse_options(self, options):
        """Parse user options and call the correct pipeline(s)"""

        if(options.subparser_name == 'de_novo_wf'):
            check_dependencies(['prodigal', 'hmmalign'])
            if (options.cpus > 1):
                check_dependencies(['FastTreeMP'])
            else:
                check_dependencies(['FastTree'])

            self.identify(options)

            options.identify_dir = options.out_dir
            options.skip_trimming = False
            self.align(options)

            if options.bac120_ms:
                options.suffix = "bac120"
            else:
                options.suffix = "ar122"

            if options.skip_gtdb_refs:
                if options.suffix == 'bac120':
                    options.msa_file = os.path.join(options.out_dir, PATH_BAC120_USER_MSA.format(prefix=options.prefix))
                elif options.suffix == 'ar122':
                    options.msa_file = os.path.join(options.out_dir, PATH_AR122_USER_MSA.format(prefix=options.prefix))
                else:
                    self.logger.error('There was an error determining the marker set.')
                    raise GenomeMarkerSetUnknown
            else:
                if options.suffix == 'bac120':
                    options.msa_file = os.path.join(options.out_dir, PATH_BAC120_MSA.format(prefix=options.prefix))
                elif options.suffix == 'ar122':
                    options.msa_file = os.path.join(options.out_dir, PATH_AR122_MSA.format(prefix=options.prefix))
                else:
                    self.logger.error('There was an error determining the marker set.')
                    raise GenomeMarkerSetUnknown

            self.infer(options)

            if options.suffix == 'bac120':
                options.input_tree = os.path.join(options.out_dir, PATH_BAC120_UNROOTED_TREE.format(prefix=options.prefix))
                options.output_tree = os.path.join(options.out_dir, PATH_BAC120_ROOTED_TREE.format(prefix=options.prefix))
            elif options.suffix == 'ar122':
                options.input_tree = os.path.join(options.out_dir, PATH_AR122_UNROOTED_TREE.format(prefix=options.prefix))
                options.output_tree = os.path.join(options.out_dir, PATH_AR122_ROOTED_TREE.format(prefix=options.prefix))
            else:
                self.logger.error('There was an error determining the marker set.')
                raise GenomeMarkerSetUnknown

            self.root(options)
            self.decorate(options)

        elif(options.subparser_name == 'classify_wf'):
            check_dependencies(
                ['prodigal', 'hmmalign', 'pplacer', 'guppy', 'fastANI'])
            self.identify(options)

            options.identify_dir = options.out_dir
            options.align_dir = options.out_dir
            options.taxa_filter = None
            options.custom_msa_filters = False
            options.skip_trimming = False  # Added here due to the other mutex argument being include above.
            options.min_consensus = None
            options.min_perc_taxa = None
            options.skip_gtdb_refs = False
            options.cols_per_gene = None
            options.max_consensus = None
            options.rnd_seed = None
            options.skip_trimming = False
            self.align(options)

            self.classify(options)
        elif (options.subparser_name == 'identify'):
            self.identify(options)
        elif(options.subparser_name == 'align'):
            self.align(options)
        elif(options.subparser_name == 'infer'):
            self.infer(options)
        elif(options.subparser_name == 'classify'):
            self.classify(options)
        elif(options.subparser_name == 'root'):
            self.root(options)
        elif(options.subparser_name == 'decorate'):
            self.decorate(options)
        elif(options.subparser_name == 'trim_msa'):
            self.trim_msa(options)
        elif(options.subparser_name == 'export_msa'):
            self.export_msa(options)
        elif(options.subparser_name == 'test'):
            self.run_test(options)
        elif(options.subparser_name == 'check_install'):
            self.check_install()
        else:
            self.logger.error('Unknown GTDB-Tk command: "' +
                              options.subparser_name + '"\n')
            sys.exit()

        return 0



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import logging

import config.config as Config

from biolib_lite.seq_io import read_fasta
from biolib_lite.common import remove_extension, make_sure_path_exists


from shutil import copyfile

from gtdbtk.exceptions import ReferenceFileMalformed


class Misc():

    def __init__(self):
        """Initialize."""
        self.logger = logging.getLogger('timestamp')

    def trim_msa(self, untrimmed_msa, mask_type, maskid, output_file):
        if maskid == 'bac' and mask_type == 'reference':
            mask = os.path.join(Config.MASK_DIR, Config.MASK_BAC120)
        elif maskid == 'arc' and mask_type == 'reference':
            mask = os.path.join(Config.MASK_DIR, Config.MASK_AR122)
        elif mask_type == 'file':
            mask = maskid
        with open(mask, 'r') as f:
            maskstr = f.readline()

        outfwriter = open(output_file, 'w')
        dict_genomes = read_fasta(untrimmed_msa, False)

        for k, v in dict_genomes.iteritems():
            aligned_seq = ''.join([v[i] for i in range(0, len(maskstr)) if maskstr[i] == '1'])
            fasta_outstr = ">%s\n%s\n" % (k, aligned_seq)
            outfwriter.write(fasta_outstr)
        outfwriter.close()
        return True

    def export_msa(self, domain, output_file):
        file_to_export = Config.CONCAT_BAC120
        if domain == 'arc':
            file_to_export = Config.CONCAT_AR122

        make_sure_path_exists(os.path.dirname(output_file))
        copyfile(file_to_export, output_file)

    def checkfile(self, file_path, file_name):
        if os.path.exists(file_path) and os.path.getsize(file_path) > 0:
            self.logger.info("Check file {} ({}): OK".format(file_name, file_path))
            return True
        else:
            self.logger.warning("Check file {} ({}): missing".format(file_name, file_path))
            return False

    def checkfolder(self, folder_path, folder_name):
        if os.path.isdir(folder_path) and len(os.listdir(folder_path)) > 0:
            self.logger.info("Check folder {} ({}): OK".format(folder_name, folder_path))
            return True
        else:
            self.logger.warning("Check folder {} ({}): missing".format(folder_name, folder_path))
            return False

    def check_install(self):
        ok = True
        ok = ok and self.checkfile(Config.TAXONOMY_FILE, 'Taxonomy')
        ok = ok and self.checkfile(Config.CONCAT_BAC120, 'concat_bac120')
        ok = ok and self.checkfile(Config.CONCAT_AR122, 'concat_ar122')
        ok = ok and self.checkfile(os.path.join(Config.MASK_DIR,  Config.MASK_BAC120), 'mask_bac120')
        ok = ok and self.checkfile(os.path.join(Config.MASK_DIR,  Config.MASK_AR122), 'mask_ar122')
        ok = ok and self.checkfile(Config.TIGRFAM_HMMS, 'tirgfam_hmms')
        pfam_test_file = os.path.join(Config.PFAM_HMM_DIR, 'Pfam-A.hmm')
        ok = ok and self.checkfile(pfam_test_file, 'pfam_hmms')

        ok = ok and self.checkfolder(Config.FASTANI_GENOMES, 'fastani_genomes')
        ok = ok and self.checkfolder(os.path.join(Config.PPLACER_DIR,  Config.PPLACER_BAC120_REF_PKG), 'pplacer_bac120')
        ok = ok and self.checkfolder(os.path.join(Config.PPLACER_DIR, Config.PPLACER_AR122_REF_PKG), 'pplacer_ar122')

        if not ok:
            self.logger.error('One or more reference files are malformed.')
            raise ReferenceFileMalformed



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from __future__ import print_function

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2014'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

import sys
import logging
import traceback
import multiprocessing as mp

import seq_io as seq_io


class Parallel(object):
    """Processes data in parallel.

    This class implements a producer/consumer model for
    processing data in parallel. In order to keep data
    handling efficient the main process acts as the
    consumer. This keeps all consumed data in the
    address space of the main process.

    Producer and consumer functions are passed to
    the run() method. The return value of the producer
    function is provided to the consumer as an argument.
    The producer function must not return None. If a generic
    return value is required simply return True. A progress
    function can also be specified which allows a progress
    string to be defined.

    Example:

        class Test(object):
            def _producer(self, data):
                return data * data

            def _consumer(self, produced_data, consumer_data):
                if consumer_data == None:
                    # setup structure for consumed data
                    consumer_data = 0

                consumer_data += produced_data

                return consumer_data

            def _progress(self, processed_items, total_items):
                return 'Processed %d of %d items.' % (processed_items, total_items)

            def run(self):
                parallel = Parallel(cpus = 2)
                consumer_data = parallel.run(self._producer,
                                                self._consumer,
                                                data_items = [1, 2, 3, 4, 5],
                                                self._progress)

                print consumer_data
    """

    def __init__(self, cpus=1):
        """Initialization.

        cpus : int
            Number of processes to create.

        """

        self.logger = logging.getLogger('timestamp')

        self.cpus = cpus

    def __producer(self, producer_callback, producer_queue, consumer_queue):
        """Process data items with producer processes.

        Parameters
        ----------
        producer_callback : function
            Function to process data items.
        producer_queue : queue
            Data items to process.
        consumer_queue : queue
            Queue for holding processed data items to be consumed.
        """
        while True:
            dataItem = producer_queue.get(block=True, timeout=None)
            if dataItem == None:
                break

            rtn = producer_callback(dataItem)

            consumer_queue.put(rtn)

    def __process_manager(self, producer_callback, producer_queue, consumer_queue):
        """Setup producer processes and manage completion of processes.

        Parameters
        ----------
        producer_callback : function
            Function to process data items.
        producer_queue : queue
            Data items to process.
        consumer_queue : queue
            Queue for holding processed data items to be consumed.
        """

        try:
            producer_proc = [mp.Process(target=self.__producer, args=(
                producer_callback, producer_queue, consumer_queue)) for _ in range(self.cpus)]

            for p in producer_proc:
                p.start()

            for p in producer_proc:
                p.join()

            consumer_queue.put(None)
        except:
            for p in producer_proc:
                p.terminate()

    def run(self, producer, consumer, data_items, progress=None):
        """Setup producers and consumer processes.

        The producer function must be specified and must
        not return None. Consumer and progress can be set to None.

        Parameters
        ----------
        producer : function
            Function to process data items.
        consumer : queue
            Function to consumed processed data items.
        data_items : list
            Data items to process.
        progress : function
            Function to report progress string.

        Returns
        -------
        <user specified>
            Set by caller in the consumer function.
        """

        # populate producer queue with data to process
        producer_queue = mp.Queue()
        for d in data_items:
            producer_queue.put(d)

        for _ in range(self.cpus):
            producer_queue.put(None)  # signal processes to terminate

        try:
            consumer_queue = mp.Queue()
            manager_proc = mp.Process(target=self.__process_manager, args=(
                producer, producer_queue, consumer_queue))

            manager_proc.start()

            # process items produced by workers
            items_processed = 0
            consumer_data = None
            while True:
                if progress:
                    status = progress(items_processed, len(data_items))
                    sys.stdout.write('%s\r' % status)
                    sys.stdout.flush()

                produced_data = consumer_queue.get(block=True, timeout=None)
                if produced_data == None:
                    break

                if consumer:
                    consumer_data = consumer(produced_data, consumer_data)

                items_processed += 1

            if progress:
                sys.stdout.write('\n')

            manager_proc.join()

            return consumer_data
        except Exception as _err:
            print(sys.exc_info()[0])
            print(traceback.format_exc())
            self.logger.warning('Exception encountered while processing data.')
            manager_proc.terminate()

    def run_seqs_file(self, producer, consumer, seq_file, progress=None):
        """Process sequences in parallel.

        The producer function must be specified and must
        not return None. Consumer and progress can be set to None.

        Parameters
        ----------
        producer : function
            Function to process data items.
        consumer : queue
            Function to consumed processed data items.
        seq_file : str
            Name of fasta/q file to read.
        progress : function
            Function to report progress string.

        Returns
        -------
        <user specified>
            Set by caller in the consumer function.
        """

        # populate producer queue with data to process
        seq_iter = seq_io.read_seq(seq_file)
        producer_queue = mp.Queue()
        read_all_seqs = False
        for _ in range(self.cpus):
            try:
                seq_data = seq_iter.next()
                producer_queue.put(seq_data)
            except StopIteration:
                read_all_seqs = True
                for _ in range(self.cpus):
                    producer_queue.put(None)  # signal processes to terminate
                break

        data_items = sum(1 for _ in seq_io.read_seq(seq_file))
        try:
            consumer_queue = mp.Queue()
            manager_proc = mp.Process(target=self.__process_manager, args=(
                producer, producer_queue, consumer_queue))

            manager_proc.start()

            # process items produced by workers
            items_processed = 0
            consumer_data = None
            while True:
                if progress:
                    status = progress(items_processed, data_items)
                    sys.stdout.write('%s\r' % status)
                    sys.stdout.flush()

                produced_data = consumer_queue.get(block=True, timeout=None)
                if produced_data == None:
                    break

                if not read_all_seqs:
                    try:
                        seq_data = seq_iter.next()
                        producer_queue.put(seq_data)
                    except StopIteration:
                        read_all_seqs = True
                        for _ in range(self.cpus):
                            # signal processes to terminate
                            producer_queue.put(None)

                if consumer:
                    consumer_data = consumer(produced_data, consumer_data)

                items_processed += 1

            if progress:
                sys.stdout.write('\n')

            manager_proc.join()

            return consumer_data
        except Exception as _err:
            print(sys.exc_info()[0])
            print(traceback.format_exc())
            self.logger.warning('Exception encountered while processing data.')
            manager_proc.terminate()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
from collections import defaultdict, namedtuple

from biolib_lite.taxonomy import Taxonomy
from biolib_lite.common import is_float
from biolib_lite.newick import parse_label

import dendropy

from numpy import (mean as np_mean,
                   std as np_std,
                   median as np_median,
                   abs as np_abs,
                   array as np_array,
                   arange as np_arange,
                   linspace as np_linspace,
                   percentile as np_percentile,
                   ones_like as np_ones_like,
                   histogram as np_histogram)


class RelativeDistance():
    """Determine relative rates of evolutionary divergence.

    This code is based on Phylorank: https://github.com/dparks1134/PhyloRank 

    """

    def __init__(self):
        """Initialization."""
        self.logger = logging.getLogger()

        #Options = namedtuple('Options', 'width height tick_font_size label_font_size dpi')
        #options = Options(6, 6, 12, 12, 96)
        #AbstractPlot.__init__(self, options)
        #self.dpi = 96

    def _avg_descendant_rate(self, tree):
        """Calculate average rate of divergence for each nodes in a tree.

        The average rate is the arithmetic mean of the
        branch length to all descendant taxa.

        Parameters
        ----------
        tree : Dendropy Tree
            Phylogenetic tree.

        Returns
        -------
        The following attributes are added to each node:
          mean_dist: mean distance to tips
          num_taxa: number of terminal taxa
        """

        # calculate the mean branch length to extant taxa
        for node in tree.postorder_node_iter():
            avg_div = 0
            if node.is_leaf():
                node.mean_dist = 0.0
                node.num_taxa = 1
            else:
                node.num_taxa = sum([1 for _ in node.leaf_iter()])
                for c in node.child_node_iter():
                    num_tips = c.num_taxa
                    avg_div += (float(c.num_taxa) / node.num_taxa) * \
                        (c.mean_dist + c.edge_length)

            node.mean_dist = avg_div

    def decorate_rel_dist(self, tree):
        """Calculate relative distance to each internal node.

        Parameters
        ----------
        tree : Dendropy Tree
            Phylogenetic tree.

        Returns
        -------
        The following attributes are added to each node:
          mean_dist: mean distance to tips
          num_taxa: number of terminal taxa
          rel_dists: relative distance of node between root and extant organisms
        """

        self._avg_descendant_rate(tree)

        for node in tree.preorder_node_iter():
            if node == tree.seed_node:
                node.rel_dist = 0.0
            elif node.is_leaf():
                node.rel_dist = 1.0
            else:
                a = node.edge_length
                b = node.mean_dist
                x = node.parent_node.rel_dist

                if (a + b) != 0:
                    rel_dist = x + (a / (a + b)) * (1.0 - x)
                else:
                    # internal node has zero length to parent,
                    # so should have the same relative distance
                    # as the parent node
                    rel_dist = x

                node.rel_dist = rel_dist

    def rel_dist_to_named_clades(self, tree):
        """Determine relative distance to specific taxa.

        Parameters
        ----------
        tree : Dendropy Tree
            Phylogenetic tree.

        Returns
        -------
        dict : d[rank_index][taxon] -> relative divergence
        """

        # calculate relative distance for all nodes
        self.decorate_rel_dist(tree)

        # assign internal nodes with ranks from
        rel_dists = defaultdict(dict)
        for node in tree.preorder_node_iter(lambda n: n != tree.seed_node):
            if not node.label or node.is_leaf():
                continue

            # check for support value
            _support, taxon_name, _auxiliary_info = parse_label(node.label)

            if not taxon_name:
                continue

            # get most-specific rank if a node represents multiple ranks
            if ';' in taxon_name:
                taxon_name = taxon_name.split(';')[-1].strip()

            most_specific_rank = taxon_name[0:3]
            rel_dists[Taxonomy.rank_index[most_specific_rank]
                      ][taxon_name] = node.rel_dist

        return rel_dists

    def _distribution_summary_plot(self, phylum_rel_dists, taxa_for_dist_inference, plot_file):
        """Summary plot showing the distribution of taxa at each taxonomic rank under different rootings.

        Parameters
        ----------
        phylum_rel_dists: phylum_rel_dists[phylum][rank_index][taxon] -> relative divergences
            Relative divergence of taxon at each rank for different phylum-level rootings.
        taxa_for_dist_inference : iterable
            Taxa to considered when inferring distributions.
        plot_file : str
            Desired name of output plot.
        """

        self.fig.clear()
        self.fig.set_size_inches(12, 6)
        ax = self.fig.add_subplot(111)

        # determine median relative distance for each taxa
        medians_for_taxa = self.taxa_median_rd(phylum_rel_dists)

        # create percentile and classification boundary lines
        percentiles = {}
        for i, rank in enumerate(sorted(medians_for_taxa.keys())):
            v = [np_median(dists) for taxon, dists in medians_for_taxa[rank].iteritems(
            ) if taxon in taxa_for_dist_inference]
            if not v:
                # not taxa at rank suitable for creating classification
                # boundaries
                continue

            p10, p50, p90 = np_percentile(v, [10, 50, 90])
            ax.plot((p10, p10), (i, i + 0.25),
                    c=(0.3, 0.3, 0.3), lw=2, zorder=2)
            ax.plot((p50, p50), (i, i + 0.5),
                    c=(0.3, 0.3, 0.3), lw=2, zorder=2)
            ax.plot((p90, p90), (i, i + 0.25),
                    c=(0.3, 0.3, 0.3), lw=2, zorder=2)

            for b in [-0.2, -0.1, 0.1, 0.2]:
                boundary = p50 + b
                if boundary < 1.0 and boundary > 0.0:
                    if abs(b) == 0.1:
                        c = (1.0, 0.65, 0.0)  # orange
                    else:
                        c = (1.0, 0.0, 0.0)
                    ax.plot((boundary, boundary),
                            (i, i + 0.5), c=c, lw=2, zorder=2)

            percentiles[i] = [p10, p50, p90]

        # create scatter plot and results table
        x = []
        y = []
        c = []
        labels = []
        rank_labels = []
        for i, rank in enumerate(sorted(medians_for_taxa.keys())):
            rank_label = Taxonomy.rank_labels[rank]
            rank_labels.append(rank_label + ' (%d)' %
                               len(medians_for_taxa[rank]))

            mono = []
            poly = []
            no_inference = []
            for clade_label, dists in medians_for_taxa[rank].iteritems():
                md = np_median(dists)
                x.append(md)
                y.append(i)
                labels.append(clade_label)

                if self._is_integer(clade_label.split('^')[-1]):
                    # taxa with a numerical suffix after a caret indicate
                    # polyphyletic groups when decorated with tax2tree
                    c.append((1.0, 0.0, 0.0))
                    poly.append(md)
                elif clade_label not in taxa_for_dist_inference:
                    c.append((0.3, 0.3, 0.3))
                    no_inference.append(md)
                else:
                    c.append((0.0, 0.0, 1.0))
                    mono.append(md)

            # histogram for each rank
            n = 0
            if len(mono) > 0:
                mono = np_array(mono)
                no_inference = np_array(no_inference)
                poly = np_array(poly)
                binwidth = 0.025
                bins = np_arange(0, 1.0 + binwidth, binwidth)

                mono_max_count = max(np_histogram(mono, bins=bins)[0])
                mono_weights = np_ones_like(mono) * (1.0 / mono_max_count)

                w = float(len(mono)) / (len(mono) +
                                        len(poly) + len(no_inference))
                n, b, p = ax.hist(mono, bins=bins,
                                  color=(0.0, 0.0, 1.0),
                                  alpha=0.25,
                                  weights=0.9 * w * mono_weights,
                                  bottom=i,
                                  lw=0,
                                  zorder=0)

            if len(no_inference) > 0:
                no_inference_max_count = max(
                    np_histogram(no_inference, bins=bins)[0])
                no_inference_weights = np_ones_like(
                    no_inference) * (1.0 / no_inference_max_count)

                ax.hist(no_inference, bins=bins,
                        color=(0.3, 0.3, 0.3),
                        alpha=0.25,
                        weights=0.9 * (1.0 - w) * no_inference_weights,
                        bottom=i + n,
                        lw=0,
                        zorder=0)

            if len(poly) > 0:
                poly_max_count = max(np_histogram(poly, bins=bins)[0])
                poly_weights = np_ones_like(poly) * (1.0 / poly_max_count)

                ax.hist(poly, bins=bins,
                        color=(1.0, 0.0, 0.0),
                        alpha=0.25,
                        weights=0.9 * (1.0 - w) * poly_weights,
                        bottom=i + n,
                        lw=0,
                        zorder=0)

        scatter = ax.scatter(x, y, alpha=0.5, s=48, c=c, zorder=1)

        # set plot elements
        ax.grid(color=(0.8, 0.8, 0.8), linestyle='dashed')

        ax.set_xlabel('relative distance')
        ax.set_xticks(np_arange(0, 1.05, 0.1))
        ax.set_xlim([-0.01, 1.01])

        ax.set_ylabel('rank (no. taxa)')
        ax.set_yticks(range(0, len(medians_for_taxa)))
        ax.set_ylim([-0.2, len(medians_for_taxa) - 0.01])
        ax.set_yticklabels(rank_labels)

        self.prettify(ax)

        # make plot interactive
        mpld3.plugins.clear(self.fig)
        mpld3.plugins.connect(
            self.fig, mpld3.plugins.PointLabelTooltip(scatter, labels=labels))
        mpld3.plugins.connect(
            self.fig, mpld3.plugins.MousePosition(fontsize=10))
        mpld3.save_html(self.fig, plot_file[0:plot_file.rfind('.')] + '.html')

        self.fig.tight_layout(pad=1)
        self.fig.savefig(plot_file, dpi=self.dpi)

    def _median_summary_outlier_file(self, phylum_rel_dists,
                                     taxa_for_dist_inference,
                                     gtdb_parent_ranks,
                                     outlier_table,
                                     rank_file,
                                     verbose_table):
        """Identify outliers relative to the median of rank distributions.
        Parameters
        ----------
        phylum_rel_dists: phylum_rel_dists[phylum][rank_index][taxon] -> relative divergences
            Relative divergence of taxon at each rank for different phylum-level rootings.
        taxa_for_dist_inference : iterable
            Taxa to considered when inferring distributions.
        gtdb_parent_ranks: d[taxon] -> string indicating parent taxa
            Parent taxa for each taxon.
        outlier_table : str
            Desired name of output table.
        rank_file : str
            Desired name of file indicating median relative distance of each rank.
        verbose_table : boolean
            Print additional columns in output table.
        """

        # determine median relative distance for each taxa
        medians_for_taxa = self.taxa_median_rd(phylum_rel_dists)

        # determine median relative distance for each rank
        median_for_rank = self.rank_median_rd(
            phylum_rel_dists, taxa_for_dist_inference)

        fout_rank = open(rank_file, 'w')
        median_str = []
        for rank in sorted(median_for_rank.keys()):
            median_str.append(
                '"' + Taxonomy.rank_labels[rank] + '":' + str(median_for_rank[rank]))
        fout_rank.write('{' + ','.join(median_str) + '}\n')
        fout_rank.close()

        fout = open(outlier_table, 'w')
        if verbose_table:
            fout.write('Taxa\tGTDB taxonomy\tMedian distance')
            fout.write('\tMedian of rank\tMedian difference')
            fout.write('\tClosest rank\tClassifciation\n')
        else:
            fout.write(
                'Taxa\tGTDB taxonomy\tMedian distance\tMedian difference\tClosest rank\tClassification\n')

        for rank in sorted(median_for_rank.keys()):
            for clade_label, dists in medians_for_taxa[rank].iteritems():
                dists = np_array(dists)

                taxon_median = np_median(dists)
                delta = taxon_median - median_for_rank[rank]

                closest_rank_dist = 1e10
                for test_rank, test_median in median_for_rank.iteritems():
                    abs_dist = abs(taxon_median - test_median)
                    if abs_dist < closest_rank_dist:
                        closest_rank_dist = abs_dist
                        closest_rank = Taxonomy.rank_labels[test_rank]

                classification = "OK"
                if delta < -0.2:
                    classification = "very overclassified"
                elif delta < -0.1:
                    classification = "overclassified"
                elif delta > 0.2:
                    classification = "very underclassified"
                elif delta > 0.1:
                    classification = "underclassified"

                if verbose_table:
                    fout.write('%s\t%s\t%.2f\t%.3f\t%.3f\t%s\t%s\n' % (clade_label,
                                                                       ';'.join(
                                                                           gtdb_parent_ranks[clade_label]),
                                                                       taxon_median,
                                                                       median_for_rank[rank],
                                                                       delta,
                                                                       closest_rank,
                                                                       classification))
                else:
                    fout.write('%s\t%s\t%.3f\t%.3f\t%s\t%s\n' % (clade_label,
                                                                 ';'.join(
                                                                     gtdb_parent_ranks[clade_label]),
                                                                 taxon_median,
                                                                 delta,
                                                                 closest_rank,
                                                                 classification))
        fout.close()

    def rank_median_rd(self, phylum_rel_dists, taxa_for_dist_inference):
        """Calculate median relative divergence for each rank.

        Parameters
        ----------
        phylum_rel_dists: phylum_rel_dists[phylum][rank_index][taxon] -> relative divergences
            Relative divergence of taxon at each rank for different phylum-level rootings.
        taxa_for_dist_inference : iterable
            Taxa to considered when inferring distributions.
        """

        medians_for_taxa = self.taxa_median_rd(phylum_rel_dists)

        median_for_rank = {}
        for i, rank in enumerate(sorted(medians_for_taxa.keys())):
            v = [np_median(dists) for taxon, dists in medians_for_taxa[rank].iteritems(
            ) if taxon in taxa_for_dist_inference]

            if v:
                median_for_rank[rank] = np_median(v)

        return median_for_rank

    def taxa_median_rd(self, phylum_rel_dists):
        """Calculate the median relative divergence for each taxon.

        Parameters
        ----------
        phylum_rel_dists: phylum_rel_dists[phylum][rank_index][taxon] -> relative divergences
            Relative divergence of taxon at each rank for different phylum-level rootings.    
        """

        medians_for_taxa = defaultdict(lambda: defaultdict(list))
        for p in phylum_rel_dists:
            for rank, d in phylum_rel_dists[p].iteritems():
                for taxon, dist in d.iteritems():
                    medians_for_taxa[rank][taxon].append(dist)

        return medians_for_taxa

    def _is_integer(self, s):
        """Test if a string represents an integer."""
        try:
            int(s)
            return True
        except ValueError:
            return False



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging
import ntpath

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from common import make_sure_path_exists


def logger_setup(log_dir, log_file, program_name, version, silent):
    """Setup loggers.

    Two logger are setup which both print to the stdout and a 
    log file when the log_dir is not None. The first logger is
    named 'timestamp' and provides a timestamp with each call,
    while the other is named 'no_timestamp' and does not prepend
    any information. The attribution 'is_silent' is also added
    to each logger to indicate if the silent flag is thrown.

    Parameters
    ----------
    log_dir : str
        Output directory for log file.
    log_file : str
        Desired name of log file.
    program_name : str
        Name of program.
    version : str
        Program version number.
    silent : boolean
        Flag indicating if output to stdout should be suppressed.
    """

    # setup general properties of loggers
    timestamp_logger = logging.getLogger('timestamp')
    timestamp_logger.setLevel(logging.DEBUG)
    log_format = logging.Formatter(fmt="[%(asctime)s] %(levelname)s: %(message)s",
                                   datefmt="%Y-%m-%d %H:%M:%S")

    no_timestamp_logger = logging.getLogger('no_timestamp')
    no_timestamp_logger.setLevel(logging.DEBUG)

    # setup logging to console
    timestamp_stream_logger = logging.StreamHandler(sys.stdout)
    timestamp_stream_logger.setFormatter(log_format)
    timestamp_logger.addHandler(timestamp_stream_logger)

    no_timestamp_stream_logger = logging.StreamHandler(sys.stdout)
    no_timestamp_stream_logger.setFormatter(None)
    no_timestamp_logger.addHandler(no_timestamp_stream_logger)

    timestamp_logger.is_silent = False
    no_timestamp_stream_logger.is_silent = False
    if silent:
        timestamp_logger.is_silent = True
        timestamp_stream_logger.setLevel(logging.ERROR)
        no_timestamp_stream_logger.is_silent = True

    if log_dir:
        make_sure_path_exists(log_dir)
        timestamp_file_logger = logging.FileHandler(
            os.path.join(log_dir, log_file), 'a')
        timestamp_file_logger.setFormatter(log_format)
        timestamp_logger.addHandler(timestamp_file_logger)

        no_timestamp_file_logger = logging.FileHandler(
            os.path.join(log_dir, log_file), 'a')
        no_timestamp_file_logger.setFormatter(None)
        no_timestamp_logger.addHandler(no_timestamp_file_logger)

    timestamp_logger.info('%s v%s' % (program_name, version))
    timestamp_logger.info(ntpath.basename(
        sys.argv[0]) + ' ' + ' '.join(sys.argv[1:]))



import os

# Command: identify
DIR_IDENTIFY = 'identify'
DIR_IDENTIFY_INTERMEDIATE = os.path.join(DIR_IDENTIFY, 'intermediate_results')
DIR_MARKER_GENE = os.path.join(DIR_IDENTIFY_INTERMEDIATE, 'marker_genes')
PATH_BAC120_MARKER_SUMMARY = os.path.join(DIR_IDENTIFY, '{prefix}.bac120.markers_summary.tsv')
PATH_AR122_MARKER_SUMMARY = os.path.join(DIR_IDENTIFY, '{prefix}.ar122.markers_summary.tsv')
PATH_TLN_TABLE_SUMMARY = os.path.join(DIR_IDENTIFY, '{prefix}.translation_table_summary.tsv')

GENOME_FILE_SUFFIX = "_genomic.fna"
PROTEIN_FILE_SUFFIX = "_protein.faa"
NT_GENE_FILE_SUFFIX = "_protein.fna"
GFF_FILE_SUFFIX = "_protein.gff"
TRANSLATION_TABLE_SUFFIX = "_translation_table.tsv"
CHECKSUM_SUFFIX = ".sha256"
TIGRFAM_SUFFIX = "_tigrfam.tsv"
TIGRFAM_TOP_HIT_SUFFIX = "_tigrfam_tophit.tsv"
PFAM_SUFFIX = "_pfam.tsv"
PFAM_TOP_HIT_SUFFIX = "_pfam_tophit.tsv"

# Command: align
DIR_ALIGN = 'align'
DIR_ALIGN_INTERMEDIATE = os.path.join(DIR_ALIGN, 'intermediate_results')
PATH_BAC120_FILTERED_GENOMES = os.path.join(DIR_ALIGN, '{prefix}.bac120.filtered.tsv')
PATH_AR122_FILTERED_GENOMES = os.path.join(DIR_ALIGN, '{prefix}.ar122.filtered.tsv')
PATH_BAC120_MSA = os.path.join(DIR_ALIGN, '{prefix}.bac120.msa.fasta')
PATH_AR122_MSA = os.path.join(DIR_ALIGN, '{prefix}.ar122.msa.fasta')
PATH_BAC120_USER_MSA = os.path.join(DIR_ALIGN, '{prefix}.bac120.user_msa.fasta')
PATH_AR122_USER_MSA = os.path.join(DIR_ALIGN, '{prefix}.ar122.user_msa.fasta')
PATH_BAC120_MARKER_INFO = os.path.join(DIR_ALIGN_INTERMEDIATE, '{prefix}.bac120.marker_info.tsv')
PATH_AR122_MARKER_INFO = os.path.join(DIR_ALIGN_INTERMEDIATE, '{prefix}.ar122.marker_info.tsv')

# Command: classify
DIR_CLASSIFY = 'classify'
PATH_BAC120_TREE_FILE = os.path.join(DIR_CLASSIFY, '{prefix}.bac120.classify.tree')
PATH_AR122_TREE_FILE = os.path.join(DIR_CLASSIFY, '{prefix}.ar122.classify.tree')
PATH_BAC120_SUMMARY_OUT = os.path.join(DIR_CLASSIFY, '{prefix}.bac120.summary.tsv')
PATH_AR122_SUMMARY_OUT = os.path.join(DIR_CLASSIFY, '{prefix}.ar122.summary.tsv')

DIR_CLASSIFY_INTERMEDIATE = os.path.join(DIR_CLASSIFY, 'intermediate_results')
PATH_BAC120_RED_DICT = os.path.join(DIR_CLASSIFY_INTERMEDIATE, '{prefix}.bac120.red_dictionary.tsv')
PATH_AR122_RED_DICT = os.path.join(DIR_CLASSIFY_INTERMEDIATE, '{prefix}.ar122.red_dictionary.tsv')
PATH_BAC120_PPLACER_CLASS = os.path.join(DIR_CLASSIFY_INTERMEDIATE, '{prefix}.bac120.classification_pplacer.tsv')
PATH_AR122_PPLACER_CLASS = os.path.join(DIR_CLASSIFY_INTERMEDIATE, '{prefix}.ar122.classification_pplacer.tsv')

DIR_PPLACER = os.path.join(DIR_CLASSIFY_INTERMEDIATE, 'pplacer')
PATH_BAC120_PPLACER_OUT = os.path.join(DIR_PPLACER, 'pplacer.bac120.out')
PATH_AR122_PPLACER_OUT = os.path.join(DIR_PPLACER, 'pplacer.ar122.out')
PATH_BAC120_PPLACER_JSON = os.path.join(DIR_PPLACER, 'pplacer.bac120.json')
PATH_AR122_PPLACER_JSON = os.path.join(DIR_PPLACER, 'pplacer.ar122.json')

# Command: infer
DIR_INFER = 'infer'
DIR_INFER_INTERMEDIATE = os.path.join(DIR_INFER, 'intermediate_results')
PATH_MARKER_UNROOTED_TREE = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.{marker}.unrooted.tree')
PATH_UNROOTED_TREE = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.unrooted.tree')
PATH_MARKER_TREE_LOG = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.{marker}.tree.log')
PATH_TREE_LOG = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.tree.log')
PATH_MARKER_FASTTREE_LOG = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.{marker}.fasttree.log')
PATH_FASTTREE_LOG = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.fasttree.log')

PATH_BAC120_ROOTED_TREE = os.path.join(DIR_INFER, '{prefix}.bac120.rooted.tree')
PATH_AR122_ROOTED_TREE = os.path.join(DIR_INFER, '{prefix}.ar122.rooted.tree')
PATH_BAC120_UNROOTED_TREE = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.bac120.unrooted.tree')
PATH_AR122_UNROOTED_TREE = os.path.join(DIR_INFER_INTERMEDIATE, '{prefix}.ar122.unrooted.tree')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import logging
import subprocess


"""Helper functions for running external programs."""


def run(cmd):
    """Run executable program.

    Parameters
    ----------
    command : str
        Command to run.

    Returns
    -------
    boolean
        True if executed , else False.
    exception
        Program output if executed successfully, else Exception.
    """

    try:
        output = subprocess.check_output(cmd.split(' '))
        return True, output
    except subprocess.CalledProcessError as e:
        return False, e

        #logging.error('Failed to execute:')
        # logging.error(e.cmd)
        # logging.error('')
        #logging.error('Program %s returned:' % args[0])
        # logging.error(e.output)


def is_executable(fpath):
    """Check if file is executable.

    This is a Python implementation of the linux
    command 'which'.

    Parameters
    ----------
    fpath : str
        Path to file.

    Returns
    -------
    boolean
        True if executable, else False.
    """

    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(program):
    """Return path to program.

    This is a Python implementation of the linux
    command 'which'.

    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python

    Parameters
    ----------
    program : str
        Name of executable for program.

    Returns
    -------
    str
        Path to executable, or None if it isn't on the path.
    """

    fpath, _fname = os.path.split(program)
    if fpath:
        if is_executable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_executable(exe_file):
                return exe_file

    return None


def check_on_path(program, exit_on_fail=True):
    """Check if program is on the system path.

    Parameters
    ----------
    program : str
        Name of executable for program.
    exit_on_fail : boolean
        Exit program with error code -1 if program in not on path.

    Returns
    -------
    boolean
        True if program is on path, else False.
    """

    if which(program):
        return True

    if exit_on_fail:
        print('%s is not on the system path.' % program)
        sys.exit(1)

    return False


def check_dependencies(programs, exit_on_fail=True):
    """Check if all required programs are on the system path.

    Parameters
    ----------
    programs : iterable
        Names of executable programs.
    exit_on_fail : boolean
        Exit program with error code -1 if any program in not on path.

    Returns
    -------
    boolean
        True if all programs are on path, else False.
    """

    for program in programs:
        if not check_on_path(program, exit_on_fail):
            return False

    return True


#!/usr/bin/env python

from setuptools import setup

import os


def version():
    setupDir = os.path.dirname(os.path.realpath(__file__))
    versionFile = open(os.path.join(setupDir, 'gtdbtk', 'VERSION'))
    return versionFile.readline().strip()


setup(
    name='gtdbtk',
    version=version(),
    author='Pierre-Alain Chaumeil and Donovan Parks',
    author_email='p.chaumeil@uq.edu.au',
    maintainer='Pierre-Alain Chaumeil and Donovan Parks',
    maintainer_email='donovan.parks@gmail.com',
    packages=['gtdbtk', 'gtdbtk.biolib_lite', 'gtdbtk.config', 'gtdbtk.external', 'gtdbtk.tests',
              'gtdbtk.external.pypfam', 'gtdbtk.external.pypfam.HMM', 'gtdbtk.external.pypfam.Scan'],
    scripts=['bin/gtdbtk'],
    package_data={'gtdbtk': ['VERSION',
                             'MANIFEST.in', 'tests/data/genomes/*']},
    url='http://pypi.python.org/pypi/gtdbtk/',
    license='GPL3',
    description='A toolkit for assigning objective taxonomic classifications to bacterial and archaeal genomes.',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    install_requires=["dendropy>=4.1.0", 'numpy', 'future'],
)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os
import subprocess

from gtdbtk.exceptions import PplacerException, TogException


class Pplacer(object):
    """ Phylogenetic placement of genomes into a tree (http://matsen.fredhutch.org/pplacer/). """

    def __init__(self):
        """ Instantiate the class. """
        self.logger = logging.getLogger('timestamp')

    def run(self, cpus, model, ref_pkg, json_out, msa_file, pplacer_out, mmap_file=None):
        """ Place genomes into the tree.

        Args:
            cpus (int): The number of threads to use.
            model (str): The model to use. Protein: LG, WAG, JTT. Nucleotides: GTR.
            ref_pkg (str): The path to the reference package.
            json_out (str): The path to write the json output to.
            msa_file (str): The path to the input MSA file.
            pplacer_out (str): Where to write the pplacer output file.
            mmap_file (str, optional): The path to write a scratch file to.

        Raises:
            PplacerException: if a non-zero exit code, or if the json output file isn't generated.

        """

        args = ['pplacer', '-m', model, '-j', str(cpus), '-c', ref_pkg, '-o', json_out, msa_file]

        if mmap_file:
            args.append('--mmap-file')
            args.append(mmap_file)

        with open(pplacer_out, 'w') as f_out:
            proc = subprocess.Popen(args, stdout=f_out, stderr=subprocess.PIPE)
            proc_out, proc_err = proc.communicate()

        if proc.returncode != 0:
            self.logger.error('An error was encountered while running pplacer.')
            raise PplacerException(proc_err)

        if not os.path.isfile(json_out):
            self.logger.error('pplacer returned a zero exit code but no output file was generated.')
            raise PplacerException

    def tog(self, pplacer_json_out, tree_file):
        """ Convert the pplacer json output into a newick tree.

        Args:
            pplacer_json_out (str): The path to the output of pplacer.
            tree_file (str): The path to output the newick file to.

        Raises:
            TogException: If a non-zero exit code is returned, or the tree file isn't output.
        """

        args = ['guppy', 'tog', '-o', tree_file, pplacer_json_out]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc_out, proc_err = proc.communicate()

        if proc.returncode != 0:
            self.logger.error('An error was encountered while running tog.')
            raise TogException(proc_err)

        if not os.path.isfile(pplacer_json_out):
            self.logger.error('tog returned a zero exit code but no output file was generated.')
            raise TogException



from __future__ import print_function

import json
import os
import sys

try:
    GENERIC_PATH = os.environ['GTDBTK_DATA_PATH']
except KeyError:
    print("'GTDBTK_DATA_PATH' environment variable is not defined")
    sys.exit(1)

############################
# If all downloaded data is in the same folder
# There is no need of editing variable below this point
############################

MIN_REF_DATA_VERSION = 'r89'

TIGRFAM_HMMS = os.path.join(GENERIC_PATH, 'markers/tigrfam/tigrfam.hmm')
PFAM_HMM_DIR = os.path.join(GENERIC_PATH, 'markers/pfam/')
MSA_FOLDER = os.path.join(GENERIC_PATH, "msa/")
MASK_DIR = os.path.join(GENERIC_PATH, "masks/")
PPLACER_DIR = os.path.join(GENERIC_PATH, "pplacer/")
FASTANI_DIR = os.path.join(GENERIC_PATH, "fastani/")
TAX_FOLDER = os.path.join(GENERIC_PATH, "taxonomy/")
RADII_DIR = os.path.join(GENERIC_PATH, "radii/")
METADATA_DIR = os.path.join(GENERIC_PATH, "metadata/")
RED_DIR = os.path.join(GENERIC_PATH, "mrca_red/")


RED_DIST_BAC_DICT = ''
RED_DIST_ARC_DICT = ''
VERSION_DATA = ''
with open(os.path.join(METADATA_DIR, "metadata.txt")) as metadataData:
    for line in metadataData:
        try:
            line_infos = line.strip().split('=')
            if line_infos[0] == 'RED_DIST_BAC_DICT':
                RED_DIST_BAC_DICT = json.loads(line_infos[1])
            elif line_infos[0] == 'RED_DIST_ARC_DICT':
                RED_DIST_ARC_DICT = json.loads(line_infos[1])
            elif line_infos[0] == 'VERSION_DATA':
                VERSION_DATA = line_infos[1]
        except ValueError:
            print("Skipping invalid line {0}".format(repr(line)))


# Relative Evolution Distance
RED_MIN_SUPPORT = 0.0
RED_MIN_CHILDREN = 2

# Marker information


BAC120_MARKERS = {"PFAM": ["PF00380.14.hmm", "PF00410.14.hmm", "PF00466.15.hmm", "PF01025.14.hmm", "PF02576.12.hmm", "PF03726.9.hmm"],
                  "TIGRFAM": ["TIGR00006.HMM", "TIGR00019.HMM", "TIGR00020.HMM", "TIGR00029.HMM", "TIGR00043.HMM", "TIGR00054.HMM", "TIGR00059.HMM", "TIGR00061.HMM", "TIGR00064.HMM",
                              "TIGR00065.HMM", "TIGR00082.HMM", "TIGR00083.HMM", "TIGR00084.HMM", "TIGR00086.HMM", "TIGR00088.HMM", "TIGR00090.HMM", "TIGR00092.HMM", "TIGR00095.HMM",
                              "TIGR00115.HMM", "TIGR00116.HMM", "TIGR00138.HMM", "TIGR00158.HMM", "TIGR00166.HMM", "TIGR00168.HMM", "TIGR00186.HMM", "TIGR00194.HMM", "TIGR00250.HMM",
                              "TIGR00337.HMM", "TIGR00344.HMM", "TIGR00362.HMM", "TIGR00382.HMM", "TIGR00392.HMM", "TIGR00396.HMM", "TIGR00398.HMM", "TIGR00414.HMM", "TIGR00416.HMM",
                              "TIGR00420.HMM", "TIGR00431.HMM", "TIGR00435.HMM", "TIGR00436.HMM", "TIGR00442.HMM", "TIGR00445.HMM", "TIGR00456.HMM", "TIGR00459.HMM", "TIGR00460.HMM",
                              "TIGR00468.HMM", "TIGR00472.HMM", "TIGR00487.HMM", "TIGR00496.HMM", "TIGR00539.HMM", "TIGR00580.HMM", "TIGR00593.HMM", "TIGR00615.HMM", "TIGR00631.HMM",
                              "TIGR00634.HMM", "TIGR00635.HMM", "TIGR00643.HMM", "TIGR00663.HMM", "TIGR00717.HMM", "TIGR00755.HMM", "TIGR00810.HMM", "TIGR00922.HMM", "TIGR00928.HMM",
                              "TIGR00959.HMM", "TIGR00963.HMM", "TIGR00964.HMM", "TIGR00967.HMM", "TIGR01009.HMM", "TIGR01011.HMM", "TIGR01017.HMM", "TIGR01021.HMM", "TIGR01029.HMM",
                              "TIGR01032.HMM", "TIGR01039.HMM", "TIGR01044.HMM", "TIGR01059.HMM", "TIGR01063.HMM", "TIGR01066.HMM", "TIGR01071.HMM", "TIGR01079.HMM", "TIGR01082.HMM",
                              "TIGR01087.HMM", "TIGR01128.HMM", "TIGR01146.HMM", "TIGR01164.HMM", "TIGR01169.HMM", "TIGR01171.HMM", "TIGR01302.HMM", "TIGR01391.HMM", "TIGR01393.HMM",
                              "TIGR01394.HMM", "TIGR01510.HMM", "TIGR01632.HMM", "TIGR01951.HMM", "TIGR01953.HMM", "TIGR02012.HMM", "TIGR02013.HMM", "TIGR02027.HMM", "TIGR02075.HMM",
                              "TIGR02191.HMM", "TIGR02273.HMM", "TIGR02350.HMM", "TIGR02386.HMM", "TIGR02397.HMM", "TIGR02432.HMM", "TIGR02729.HMM", "TIGR03263.HMM", "TIGR03594.HMM",
                              "TIGR03625.HMM", "TIGR03632.HMM", "TIGR03654.HMM", "TIGR03723.HMM", "TIGR03725.HMM", "TIGR03953.HMM"]}

AR122_MARKERS = {"PFAM": ["PF01868.11.hmm", "PF01282.14.hmm", "PF01655.13.hmm", "PF01092.14.hmm", "PF01000.21.hmm", "PF00368.13.hmm", "PF00827.12.hmm", "PF01269.12.hmm", "PF00466.15.hmm",
                          "PF01015.13.hmm", "PF13685.1.hmm", "PF02978.14.hmm", "PF04919.7.hmm", "PF01984.15.hmm", "PF04104.9.hmm", "PF00410.14.hmm", "PF01798.13.hmm", "PF01864.12.hmm",
                          "PF01990.12.hmm", "PF07541.7.hmm", "PF04019.7.hmm", "PF00900.15.hmm", "PF01090.14.hmm", "PF02006.11.hmm", "PF01157.13.hmm", "PF01191.14.hmm", "PF01866.12.hmm",
                          "PF01198.14.hmm", "PF01496.14.hmm", "PF00687.16.hmm", "PF03874.11.hmm", "PF01194.12.hmm", "PF01200.13.hmm", "PF13656.1.hmm", "PF01280.15.hmm"],
                 "TIGRFAM": ["TIGR00468.HMM", "TIGR01060.HMM", "TIGR03627.HMM", "TIGR01020.HMM", "TIGR02258.HMM", "TIGR00293.HMM", "TIGR00389.HMM", "TIGR01012.HMM", "TIGR00490.HMM", "TIGR03677.HMM",
                             "TIGR03636.HMM", "TIGR03722.HMM", "TIGR00458.HMM", "TIGR00291.HMM", "TIGR00670.HMM", "TIGR00064.HMM", "TIGR03629.HMM", "TIGR00021.HMM", "TIGR03672.HMM", "TIGR00111.HMM",
                             "TIGR03684.HMM", "TIGR01077.HMM", "TIGR01213.HMM", "TIGR01080.HMM", "TIGR00501.HMM", "TIGR00729.HMM", "TIGR01038.HMM", "TIGR00270.HMM", "TIGR03628.HMM", "TIGR01028.HMM",
                             "TIGR00521.HMM", "TIGR03671.HMM", "TIGR00240.HMM", "TIGR02390.HMM", "TIGR02338.HMM", "TIGR00037.HMM", "TIGR02076.HMM", "TIGR00335.HMM", "TIGR01025.HMM", "TIGR00471.HMM",
                             "TIGR00336.HMM", "TIGR00522.HMM", "TIGR02153.HMM", "TIGR02651.HMM", "TIGR03674.HMM", "TIGR00323.HMM", "TIGR00134.HMM", "TIGR02236.HMM", "TIGR03683.HMM", "TIGR00491.HMM",
                             "TIGR00658.HMM", "TIGR03680.HMM", "TIGR00392.HMM", "TIGR00422.HMM", "TIGR00279.HMM", "TIGR01052.HMM", "TIGR00442.HMM", "TIGR00308.HMM", "TIGR00398.HMM", "TIGR00456.HMM",
                             "TIGR00549.HMM", "TIGR00408.HMM", "TIGR00432.HMM", "TIGR00264.HMM", "TIGR00982.HMM", "TIGR00324.HMM", "TIGR01952.HMM", "TIGR03626.HMM", "TIGR03670.HMM", "TIGR00337.HMM",
                             "TIGR01046.HMM", "TIGR01018.HMM", "TIGR00936.HMM", "TIGR00463.HMM", "TIGR01309.HMM", "TIGR03653.HMM", "TIGR00042.HMM", "TIGR02389.HMM", "TIGR00307.HMM", "TIGR03673.HMM",
                             "TIGR00373.HMM", "TIGR01008.HMM", "TIGR00283.HMM", "TIGR00425.HMM", "TIGR00405.HMM", "TIGR03665.HMM", "TIGR00448.HMM"]}

RPS23_MARKERS = {"PFAM": ["PF00687.16.hmm", "PF00466.15.hmm", "PF00298.14.hmm", "PF03946.9.hmm", "PF00238.14.hmm",
                          "PF00252.13.hmm", "PF00861.17.hmm", "PF00181.18.hmm", "PF00237.14.hmm",
                          "PF03947.13.hmm", "PF00297.17.hmm", "PF00573.17.hmm", "PF00281.14.hmm",
                          "PF00673.16.hmm", "PF00338.17.hmm", "PF00411.14.hmm", "PF00416.17.hmm",
                          "PF00312.17.hmm", "PF00366.15.hmm", "PF00203.16.hmm", "PF00189.15.hmm",
                          "PF00333.15.hmm", "PF03719.10.hmm", "PF00177.16.hmm", "PF00410.14.hmm",
                          "PF00380.14.hmm", "PF00164.20.hmm"],
                 "TIGRFAM": []}

# Information for Multiple hits markers:
DEFAULT_MULTIHIT_THRESHOLD = 10.0

# Information for aligning genomes
DEFAULT_DOMAIN_THRESHOLD = 10.0
AR_MARKER_COUNT = 122
BAC_MARKER_COUNT = 120


# MSA file names
CONCAT_BAC120 = os.path.join(
    MSA_FOLDER, "gtdb_" + VERSION_DATA + "_bac120.faa")
CONCAT_AR122 = os.path.join(MSA_FOLDER, "gtdb_" + VERSION_DATA + "_ar122.faa")

# Taxonomy file name
TAXONOMY_FILE = os.path.join(TAX_FOLDER, "gtdb_taxonomy.tsv")

# Type Strain radii file
RADII_FILE = os.path.join(RADII_DIR, "gtdb_radii.tsv")

# Mask file names
MASK_BAC120 = "gtdb_" + VERSION_DATA + "_bac120.mask"
MASK_AR122 = "gtdb_" + VERSION_DATA + "_ar122.mask"
MASK_RPS23 = "gtdb_" + VERSION_DATA + "_rps23.mask"


# Pplacer configuration
PPLACER_BAC120_REF_PKG = "gtdb_" + VERSION_DATA + "_bac120.refpkg"
PPLACER_AR122_REF_PKG = "gtdb_" + VERSION_DATA + "_ar122.refpkg"
PPLACER_RPS23_REF_PKG = "gtdb_" + VERSION_DATA + "_rps23.refpkg"

# Fastani configuration
FASTANI_SPECIES_THRESHOLD = 95.0
FASTANI_GENOMES = os.path.join(FASTANI_DIR, "database/")
FASTANI_GENOMES_EXT = "_genomic.fna.gz"

# MRCA RED VALUE
MRCA_RED_BAC120 = os.path.join(
    RED_DIR, "gtdbtk_{}_bac120.tsv".format(VERSION_DATA))
MRCA_RED_AR122 = os.path.join(
    RED_DIR, "gtdbtk_{}_ar122.tsv".format(VERSION_DATA))
